@@ -15,10 +15,6 @@ G_BEGIN_DECLS
 
 typedef enum {
 	SOUP_MESSAGE_STARTING,
-	SOUP_MESSAGE_RESOLVING_PROXY_URI,
-	SOUP_MESSAGE_RESOLVED_PROXY_URI,
-	SOUP_MESSAGE_RESOLVING_PROXY_ADDRESS,
-	SOUP_MESSAGE_RESOLVED_PROXY_ADDRESS,
 	SOUP_MESSAGE_AWAITING_CONNECTION,
 	SOUP_MESSAGE_GOT_CONNECTION,
 	SOUP_MESSAGE_CONNECTING,
@@ -45,8 +41,10 @@ struct _SoupMessageQueueItem {
 	SoupAddress *proxy_addr;
 	SoupURI *proxy_uri;
 	SoupConnection *conn;
+	GSimpleAsyncResult *result;
 
 	guint paused            : 1;
+	guint new_api           : 1;
 	guint redirection_count : 31;
 
 	SoupMessageQueueItemState state;
